// POV-Ray 3.7 include file "complex.inc"
//
// Desc: POV-Ray has no complex numbers, hence this include file.
//       I chose to use a 2D vector to represent it: <Real, Imaginary>.
// Author: Ingo Janssen
// Date:  2022-07-28
// Version: 0.9 beta
//
// Assembled from various sources. Results tested against 
// https://www.wolframalpha.com/input?i=
//

#version 3.7;

#ifndef( Complex_Inc_Temp)
#declare Complex_Inc_Temp = version;

//====
// Cstr(Comp): print formatting <real, imaginary j>
//
#macro Cstr(Comp)
  concat("<",vstr(2, Comp,", ",0,-1),"j>")
#end

#macro CdebugArr(data)
  #for(i,0, dimension_size(data, 1)-1)
    #debug concat(Cstr(data[i]), "\n")
  #end
#end
  
#macro Cdebug2Darr(data)
  #for(row,0,dimension_size(data,1)-1)
    #for(col,0,dimension_size(data,1)-1)    
      #debug concat(Cstr(data[row][col]), " ")  
    #end
    #debug "\n" 
  #end 
#end

//====
// Cim, imaginary unit 'vector' <0,1j>
// Cre, real unit 'vector' <1,0>
// Complex(Raeal, Im): returns a complex number from a real and imaginary part
// I2C(Im): creates <0, Ij> complex number
// R2C(Re): creates <R, 0j> complex number
// Cr(Comp): get the real part of a complex number
// Ci(Comp): get the imaginary part of a complex number 
//
#declare Cim = <0,1>;
#declare Cre = <1,0>;
#macro Complex(Real, Im) <Real, Im> #end
#macro I2C(Im) <0, Im> #end
#macro R2C(Real) <Real, 0> #end
#macro Cr(Comp) Comp.x #end
#macro Ci(Comp) Comp.y #end

//====
// Ccompare(C1, C2): Compare two complex numbers for equality
// Cabs(Comp): Length of the <real, imaginary> vector.
// Cabs2(Comp): Squared magnitude, Cabs(Comp)^2.
// Conjugate(Comp): complex conjugate
// Cneg(Comp): negate a complex number
// Cinv(Comp): multiplicative inverse of complex number
#macro Ccompare(C1, C2) bitwise_and((C1.x = C2.x), (C1.y = C2.y)) #end
#macro Cabs(Comp) vlength(Comp) #end
#macro Cabs2(Comp) vdot(Comp, Comp) #end
#macro Conjugate(Comp) <Comp.x, -Comp.y> #end
#macro Cneg(Comp) <-Comp.x, -Comp.y> #end
#macro Cinv(Comp) Conjugate(Comp) / Cabs2(Comp) #end

//====
// RmultC(Real, Comp): Real * Comp
// CmultR(Comp, Real): Comp * Real
// CmultC(C1, C2): Comp * Comp
#macro RmultC(Real, Comp) Real * Comp #end
#macro CmultR(Comp, Real) Comp * Real #end
#macro CmultC(C1, C2) <C1.x * C2.x - C1.y * C2.y, C1.y * C2.x + C1.x * C2.y>#end

//====
// CdivR(Comp, Real): Comp / Real
// CdivC(C1, C2): Comp / Comp
// RdivC(Real, Comp): Real / Comp
#macro RdivC(Real, Comp) Real.x * Cinv(Comp) #end
#macro CdivR(Comp, Real) Comp / Real #end
#macro CdivC(C1, C2)
  #if(abs(C2.x) < abs(C2.y))
    #local R = C2.x / C2.y;
    #local D = C2.y + R * C2.x;
    #local Return = <(C1.x * R + C1.y) / D, (C1.y * R - C1.x) / D>;
  #else
    #local R = C2.y / C2.x;
    #local D = C2.x + R * C2.y;
    #local Return = <(C1.x + R * C1.y) / D, (C1.y - R * C1.x) / D>;  
  #end
  Return
#end

//====
// RaddC(Real, Comp): Real + Comp
// CaddR(Comp, Real): Comp + Real
// CaddC(C1, C2): Comp + Comp
#macro RaddC(Real, Comp) <Real.x + Comp.x, Comp.y> #end
#macro CaddR(Comp, Real) <Comp.x + Real.x, Comp.y> #end
#macro CaddC(C1, C2) C1 + C2 #end

//====
// CsubC(R, C2): substarct a complex numbers from a real
// CsubR(Comp, Real): substarct a real from complex number
// CsubC(C1, C2): substarct two complex numbers 
#macro RsubC(Real, Comp) <Real.x - Comp.x, -Comp.y>#end
#macro CsubR(Comp, Real) <Comp.x - Real.x, Comp.y>#end
#macro CsubC(C1, C2) C1 - C2 #end

//====
// Csqrt(Comp) : Principal square root of a complex number
#macro Csqrt(Comp)
  #if(Comp.x = 0 & Comp.y = 0)
    #local Return = Comp;
  #else
    #local X = abs(Comp.x);
    #local Y = abs(Comp.y);
    #if(X>=Y)
      #local R = Y / X;
      #local W = sqrt(X) * sqrt(0.5 * (1 + sqrt(1 + R * R)));
    #else
      #local R = X / Y;
      #local W = sqrt(Y) * sqrt(0.5 * (R + sqrt(1 + R * R)));    
    #end
    #if(Comp.x >= 0)
      #local Return = <W, Comp.y / (W * 2)>;
    #else 
      #local ReturnIM = (Comp.y >= 0 ? W: -W);
      #local Return = <Comp.y / (ReturnIM + ReturnIM), ReturnIM>;
    #end
  #end
  Return
#end

//====
// CpowC(C1, C2): Comp^Comp
//
#macro CpowC(C1, C2)
  #if(C1.x = 0 & C1.y = 0)
    #if(C2.x = 0 & C2.y = 0)
      #local Return = <1,0>;
    #else
      #local Return = <0,0>;
    #end
  #elseif(C2.x = 1 & C2.y = 0)
    #local Return = C1;
  #elseif(C2.x = -1 & C2.y = 0)
    #local Return = RdivC(1, C1);
  #else
    #local Rho = Cabs(C1);
    #local Theta = atan2(C1.y, C1.x);
    #local S = pow(Rho, C2.x) * exp(-C2.y * Theta);
    #local R = C2.x * Theta + C2.y * ln(Rho);
    #local Return = < S * cos(R), S * sin(R)>;
  #end
  Return
#end

//====
// CpowR(Comp, Real): Comp^Real
// Cexp(Comp): e^Comp
#macro CpowR(Comp, Real) CpowC(Comp, R2C(Real)) #end
#macro Cexp(Comp)
  #local Rho =  exp(Comp.x);
  <Rho * cos(Comp.y), Rho * sin(Comp.y)>   
#end

//====
// Cln(Comp): natural logarithm of a complex number
// Clog10(Comp): base 10 logarithm of a complex number
// Clog2(Comp): base 2 logarithm of a complex number
#macro Cln(Comp) <ln(Cabs(Comp)), atan2(Comp.y, Comp.x)> #end
#macro Clog10(Comp) Cln(Comp) / ln(10.0) #end
#macro Clog2(Comp) Cln(Comp) / ln(2.0)#end

//===
// Csin(Comp): sine of complex number
// Ccos(Comp): cosine of complex number
// Ctan(Comp): tangent of complex number
#macro Csin(Comp) <sin(Comp.x) * cosh(Comp.y), cos(Comp.x) * sinh(Comp.y)> #end
#macro Ccos(Comp) <cos(Comp.x) * cosh(Comp.y), -sin(Comp.x) * sinh(Comp.y)> #end
#macro Ctan(Comp) CdivC(Csin(Comp), Ccos(Comp)) #end

//===
// Carcsin(Comp): inverse sine of complex number
// Carccos(Comp): inverse cosine of complex number
// Carctan(Comp): inverse tangent of complex number
#macro Carcsin(Comp) CmultC(-Cim, Cln(CaddC(CmultC(Cim, Comp), Csqrt(RsubC(1, CmultC(Comp, Comp)))))) #end
#macro Carccos(Comp) CmultC(-Cim, Cln(CaddC(Comp, Csqrt(CsubR(CmultC(Comp, Comp),1))))) #end
#macro Carctan(Comp) CmultC(<0, 0.5>, CsubC(Cln(RsubC(1, CmultC(Cim, Comp))), Cln(RaddC(1, CmultC(Cim, Comp))))) #end 

//===
// Csec(Comp): secant of complex number
// Ccsc(Comp): cosecant of complex number
// Ccot(Comp): cotangent of complex number
#macro Csec(Comp) RdivC(1, Ccos(Comp)) #end 
#macro Ccsc(Comp) RdivC(1, Csin(Comp)) #end 
#macro Ccot(Comp) CdivC(Ccos(Comp), Csin(Comp))#end 

//===
// Carcsec(Comp): inverse secant of complex number
// Carccsc(Comp): inverse cosecant of complex number
// Carccot(Comp): inverse cotangent of complex number
#macro Carcsec(Comp) CmultC(-Cim, Cln(CaddC(CmultC(Cim, Csqrt(RsubC(1,RdivC(1,CmultC(Comp,Comp))))), RdivC(1,Comp)))) #end 
#macro Carccsc(Comp) CmultC(Cneg(Cim), Cln(CaddC(Csqrt(RsubC(1, RdivC(1, CmultC(Comp,Comp)))), CdivC(Cim,Comp)))) #end 
#macro Carccot(Comp) CmultC(<0, 0.5>, CsubC(Cln(RsubC(1, CdivC(Cim, Comp))), Cln(RaddC(1, CdivC(Cim, Comp))))) #end 

//===
// Csinh(Comp): hyperbolic sine of complex number
// Ccosh(Comp): hyperbolic cosine of complex number
// Ctanh(Comp): hyperbolic tangent of complex number
#macro Csinh(Comp) RmultC(0.5, CsubC(Cexp(Comp), Cexp(Cneg(Comp))))#end
#macro Ccosh(Comp) RmultC(0.5, CaddC(Cexp(Comp), Cexp(Cneg(Comp)))) #end
#macro Ctanh(Comp) CdivC(Csinh(Comp), Ccosh(Comp))#end

//===
// Carcsinh(Comp): inverse hyperbolic sine of complex number
// Carccosh(Comp): inverse hyperbolic cosine of complex number
// Carctanh(Comp): inverse hyperbolic tangent of complex number
#macro Carcsinh(Comp) Cln(CaddC(Comp, Csqrt(CaddR(CmultC(Comp,Comp),1)))) #end
#macro Carccosh(Comp) Cln(CaddC(Comp, Csqrt(CsubR(CmultC(Comp,Comp),1)))) #end
#macro Carctanh(Comp) RmultC(0.5,Cln(CdivC(RaddC(1,Comp), RsubC(1,Comp)))) #end

//===
// Csech(Comp): hyperbolic secant of complex number
// Ccsch(Comp): hyperbolic cosecant of complex number
// Ccoth(Comp): hyperbolic cotangent of complex number
#macro Csech(Comp)   RdivC(2, CaddC(Cexp(Comp), Cexp(-Comp)))#end
#macro Ccsch(Comp) RdivC(2, CsubC(Cexp(Comp), Cexp(-Comp)))#end
#macro Ccoth(Comp) CdivC(Ccosh(Comp),Csinh(Comp)) #end

//===
// Carcsech(Comp): inverse hyperbolic secant of complex number
// Carccsch(Comp): inverse hyperbolic cosecant of complex number
// Carccoth(Comp): inverse hyperbolic cotangent of complex number
#macro Carcsech(Comp) Cln(CaddC(RdivC(1,Comp), CmultC(Csqrt(CaddR(RdivC(1, Comp), 1)),Csqrt(CsubR(RdivC(1, Comp), 1))))) #end
#macro Carccsch(Comp)Cln(CaddC(RdivC(1,Comp), Csqrt(CaddR(RdivC(1,CmultC(Comp,Comp)),1)))) #end
#macro Carccoth(Comp)RmultC(0.5, CsubC(Cln(RaddC(1,RdivC(1,Comp))), Cln(RsubC(1,RdivC(1,Comp)))))#end

//===
// Cphase(Comp): angle between real and imaginary part.
// Cpolar(Comp): polar coordinates from complex number. Results in a 2d vector <r, phi>
// Crect(r, phi): complex number from r and phi. Results in a complex 'vector' <real, im> of Cartesian form.
#macro Cphase(Comp) atan2(Comp.x, Comp.y).x #end
#macro Cpolar(Comp) <Cabs(Comp), Cphase(Comp)> #end 
#macro Crect(r, phi) <RmultC(r, cos(phi)), RmultC(r, sin(phi))> #end

#if(input_file_name="complex.inc")
  global_settings{ assumed_gamma 1.0 }
  #default{ finish{ ambient 1 diffuse 0 emission 0}}

  #debug "\n\n\nRunning all complex number tests:\n\n"

  #debug concat("Complex(4,4) = (4, 4j) => ", Cstr(Complex(4,4)),"\n")
  #debug concat("I2C(4)  = (0, 4j) => ", Cstr(I2C(4)), "\n")
  #debug concat("R2C(4)  = (4, 0j) => ", Cstr(R2C(4)), "\n")
  #debug concat("Ci(<0,4>)  = 4 => ", str(Ci(<0,4>),0,-1), "\n")
  #debug concat("Cr(<4,0>)  = 4 => ", str(Cr(<4,0>),0,-1), "\n\n")
  
  #debug concat("Ccompare(<0,4>,<4,0>) = 0 => ", str(Ccompare(<0,4>,<4,0>),0,-1),"\n")
  #debug concat("Ccompare(<4,4>,<4,4>) = 1 => ", str(Ccompare(<4,4>,<4,4>),0,-1),"\n")
  #debug concat("Cabs(<1,1>) = 1.414214 => ", str(Cabs(<1,1>),0,-1),"\n")
  #debug concat("Cabs2(<1,1>) = 2.0 => ", str(Cabs2(<1,1>),0,-1),"\n")
  #debug concat("Conjugate(<1,1>) = (1, -1j) => ", Cstr(Conjugate(<1,1>)),"\n")
  #debug concat("Cneg(<1,1>) = (-1, -1j) => ", Cstr(Cneg(<1,1>)),"\n")
  #debug concat("Cinv(<1,1>) = (0.5, -0.5j) => ", Cstr(Cinv(<1,1>)),"\n\n")
  
  #debug concat("RmultC(1.5,<4,2>) = (6, 3j) => ", Cstr(RmultC(1.5, <4,2>)),"\n")
  #debug concat("CmultR(<4,2>,1.5) = (6, 3j) => ", Cstr(CmultR(<4,2>, 1.5)),"\n")
  #debug concat("CmultC(<4,2>,<1.5,1.5>) = (3, 9j) => ", Cstr(CmultC(<4,2>, <1.5,1.5>)),"\n\n")

  #debug concat("RdivC(1.5,<4,2>) = (0.3, -0.15) => ", Cstr(RdivC(1.5, <4,2>)),"\n")
  #debug concat("CdivR(<4,2>,1.5) = (2.6667, 1.3333) => ", Cstr(CdivR(<4,2>, 1.5)),"\n")
  #debug concat("CdivC(<4,2>,<1.5,1.5>) = (2.0, -0.667) => ", Cstr(CdivC(<4,2>, <1.5,1.5>)),"\n\n")

  #debug concat("RaddC(1.5,<4,2>) = (5.5, 2j) => ", Cstr(RaddC(1.5, <4,2>)),"\n")
  #debug concat("CaddR(<4,2>,1.5) = (5.5, 2j) => ", Cstr(CaddR(<4,2>, 1.5)),"\n")
  #debug concat("CaddC(<4,2>,<1.5,1.5>) = (5.5, 3.5j) => ", Cstr(CaddC(<4,2>, <1.5,1.5>)),"\n\n")

  #debug concat("RsubC(1.5,<4,2>) = (-2.5, 2j) => ", Cstr(RsubC(1.5, <4,2>)),"\n")
  #debug concat("CsubR(<4,2>,1.5) = (2.5, 2j) => ", Cstr(CsubR(<4,2>, 1.5)),"\n")
  #debug concat("CsubC(<4,2>,<1.5,1.5>) = (2.5, 0.5j) => ", Cstr(CsubC(<4,2>, <1.5,1.5>)),"\n\n")
  
  #debug concat("Csqrt(<1,1>) = (1.098684, 0.455090) => ", Cstr(Csqrt(<1,1>)),"\n")
  #debug concat("CpowC(<2,2j>, <2,2j) =(-1.452505,-0.809890i) => ", Cstr(CpowC(<2,2>, <2,2>)), "\n")
  #debug concat("CpowR(<1,1>,2) = (0, 2i) => ", Cstr(CpowR(<1,1>, 2)),"\n")
  #debug concat("Cexp(<1,1>) = (1.468694, 2.287355) => ", Cstr(Cexp(<1,1>)),"\n\n")

  #debug concat("Cln(<1,1>) = (0.346574, 0.785398) => ", Cstr(Cln(<1,1>)),"\n")
  #debug concat("Clog10(<1,1>) = (0.150515, 0.341094) => ", Cstr(Clog10(<1,1>)),"\n")
  #debug concat("Clog2(<1,1>) = (0.5, 1.133090) => ", Cstr(Clog2(<1,1>)),"\n\n")

  #debug concat("Csin(<4,4j>) = (-20.666939,-17.837880) => ", Cstr(Csin(<4,4>)), "\n")
  #debug concat("Ccos(<4,4j>) = (-17.849852, 20.653077) => ", Cstr(Ccos(<4,4>)), "\n")
  #debug concat("Ctan(<1,1j>) = (  0.271753,  1.083923) => ", Cstr(Ctan(<1,1>)), "\n\n")

  #debug concat("Carcsin(<4,4j>) = (0.777587, 2.426107) => ", Cstr(Carcsin(<4,4>)), "\n")
  #debug concat("Carccos(<4,4j>) = (0.793209,-2.426107) => ", Cstr(Carccos(<4,4>)), "\n")
  #debug concat("Carctan(<1,1j>) = (1.017222, 0.402359) => ", Cstr(Carctan(<1,1>)), "\n\n")

  #debug concat("Csec(<4,4j>) = (-0.0239542,-0.027716) => ", Cstr(Csec(<4,4>)), "\n")
  #debug concat("Ccsc(<4,4j>) = (-0.027729, 0.023933) => ", Cstr(Ccsc(<4,4>)), "\n")
  #debug concat("Ccot(<4,4j>) = (0.000664, -0.999902) => ", Cstr(Ccot(<4,4>)), "\n\n")

  #debug concat("Carcsec(<4,4j>) = (1.446456, 0.125642) => ", Cstr(Carcsec(<4,4>)), "\n")
  #debug concat("Carccsc(<4,4j>) = (0.124340, 0.125642) => ", Cstr(Carccsc(<4,4>)), "\n")
  #debug concat("Carccot(<4,4j>) = (0.126277, 0.123674) => ", Cstr(Carccot(<4,4>)), "\n\n")

  #debug concat("Csinh(<4,4j>) = (-17.8378803,-20.6669388) => ", Cstr(Csinh(<4,4>)), "\n")
  #debug concat("Ccosh(<4,4j>) = (-17.8498522,-20.6530774) => ", Cstr(Ccosh(<4,4>)), "\n")
  #debug concat("Ctanh(<4,4j>) = (1.0000974, 0.0006639) => ", Cstr(Ctanh(<4,4>)), "\n\n")

  #debug concat("Carcsinh(<4,4j>) = (2.426107, 0.777587) => ", Cstr(Carcsinh(<4,4>)), "\n")
  #debug concat("Carccosh(<4,4j>) = (2.426107, 0.793209) => ", Cstr(Carccosh(<4,4>)), "\n")
  #debug concat("Carctanh(<4,4j>) = (0.123674, 1.444519) => ", Cstr(Carctanh(<4,4>)), "\n\n")

  #debug concat("Csech(<4,4j>) = (-0.023954, 0.027716) => ", Cstr(Csech(<4,4>)), "\n")
  #debug concat("Ccsch(<4,4j>) = (-0.023933, 0.027729) => ", Cstr(Ccsch(<4,4>)), "\n")
  #debug concat("Ccoth(<4,4j>) = (0.999902,-0.000664) => ", Cstr(Ccoth(<4,4>)), "\n\n")

  #debug concat("Carcsech(<4,4j>) = (0.125642,-1.446456) => ", Cstr(Carcsech(<4,4>)), "\n")
  #debug concat("Carccsch(<4,4j>) = (0.125642,-0.124340) => ", Cstr(Carccsch(<4,4>)), "\n")
  #debug concat("Carccoth(<4,4j>) = (0.123674,-0.126277) => ", Cstr(Carccoth(<4,4>)), "\n\n")

  #debug concat("Cphase(<1,1j>) = 0.785398 => ", str(Cphase(<1,1>),0,-1), "\n")
  #debug concat("Cpolar(<1,1j>) = (r: 1.414214, phi: 0.785398) => " , "<", vstr(2,Cpolar(<1,1>), ", ", 0, -1), ">\n")
  #debug concat("Crect(r: 1.414214, phi: 0.785398) = (1.0, 1.0j) => ", Cstr(Crect(1.414214, 0.785398)), "\n\n") 

  #debug "End of complex number tests.\n\n\n" 
#end

#version Complex_Inc_Temp;
#end
