// POV-Ray 3.7 include file "fft.inc"
//
// Desc: 1D & 2D Fast Fouried Transform using the Stockham algorithm 
// Author: Ingo Janssen
// Date:  2022-07-22
// Version: 0.5 beta
//
// Requires complex.inc as POV-Ray has no complex numbers.
// Inspiration: http://wwwa.pikara.ne.jp/okojisan/otfft-en/optimization1.html 
//
// Can be optimised further by 6 & 8 step version as on the site above. 
// Also by precalculation of Theta / Wp twiddle. Separate .inc file for results?
// 
//====

#version 3.7;

#ifndef( FFT_Inc_Temp)
#declare FFT_Inc_Temp = version;
#include "complex.inc"


//====
// PowOf2(X)
//
// Determines whether X is a power of two, returns 1 if true else 0
//
// X: int: value to test
//====
#macro PowOf2(X)
  bitwise_and((X > 0), (bitwise_and(X, (X - 1)) = 0))
#end      

#macro Index2Freq(Index, SampleRate, nFFT)
  Index * (SampleRate / nFFT / 2)  
#end

#macro Freq2Index(Freq, SampleRate, nFFT)
  Freq / (SampleRate / nFFT / 2)
#end

#macro _FFT0(X, Y, N, Stride, EO)
  #local M = div(N, 2);
  #local Theta = 2 * pi / N;
  #if(N = 1)
    #if(EO)
      #for(Q, 0, Stride-1)
        #local Y[Q] = X[Q];
      #end
    #end
  #else
    #for(P, 0, M-1)
      #local Fp = P * Theta;
      #local Wp = Complex(cos(Fp), -sin(Fp)); //create and get from twiddle?
      #for(Q, 0, Stride-1)
        #local A = X[Q + Stride * (P + 0)];        
        #local B = X[Q + Stride * (P + M)];
        #local Y[Q + Stride * (2 * P + 0)] = A + B;
        #local Y[Q + Stride * (2 * P + 1)] = CmultC((A-B), Wp);
      #end
    #end
    _FFT0(Y, X, div(N, 2), 2 * Stride, !EO)
  #end
#end

 
//====
// macro FFT(X): fast Fourier transform of X
//
// X: array of complex numbers. Has to be of 'power of two length'
//
//====
#macro FFT(X)
  #local N = dimension_size(X, 1);
  #if(PowOf2(N)=0)                                   //?? Needed?
    #error "length of input is not a power of two"
  #end
  #local Y = array[N];
  _FFT0(X, Y, N, 1, false)
  #undef Y
  //#for(i,0,N-1)
    //#local X[i] = X[i]/N;
  //#end
#end


//====
// macro IFFT(X)
// inverse fast Fourier transform of X
//
// X: array of complex numbers. Has to be of 'power of two length'
//
//====
#macro IFFT(X)
  #local N = dimension_size(X,1);
  #local Fn = R2C(1/N);
  #for(P, 0, N-1)
    #local X[P] = Conjugate(CmultC(X[P],Fn));
  #end
  #local Y = array[N];
  _FFT0(X, Y, N, 1, false)
  #undef Y
  #for(P, 0, N-1)
    #local X[P] = Conjugate(X[P]);
  #end
#end


#macro FFT2D(X)
  //in place transform rows
  #for(row, 0, dimension_size(X, 1)-1)
    FFT(X[row])
  #end
  //in place transform columns
  #for(i, 0, dimension_size(X[0], 1)-1)
    #local Col = array[dimension_size(X, 1)];
    #for(j, 0, dimension_size(X, 1)-1)
      #local Col[j] = X[j][i];
    #end
    FFT(Col)
    #for(j, 0, dimension_size(X,1)-1)
      #local X[j][i] = Col[j];
    #end    
  #end
#end  


#macro IFFT2D(X)
  //in place inverse transform rows
  #for(row, 0, dimension_size(X, 1)-1)
    IFFT(X[row])
  #end
  //in place inverse transform columns
  #for(i, 0, dimension_size(X[0], 1)-1)
    #local Col = array[dimension_size(X, 1)];
    #for(j, 0, dimension_size(X, 1)-1)
      #local Col[j] = X[j][i];
    #end
    IFFT(Col)
    #for(j, 0, dimension_size(X,1)-1)
      #local X[j][i] = Col[j];
    #end    
  #end
#end  


#if(input_file_name="fft.inc")
  //
  // cmd: +H0 +W0 -D -F
  //
  // Test scene renders nothing prints only.
  // Data in -> FFT -> IFFT -> data out, should result in data out == data in.
 
  global_settings{ assumed_gamma 1.0 }
  #default{ finish{ ambient 1 diffuse 0 emission 0}}


  // #declare data = array[8]{1.0,1.0,1.0,1.0,0.0,0.0,0.0,0.0}; 
  #declare data = array[16]{0.672957, -0.453061, -0.835088, 0.980334, 0.972232, 0.640295, 0.791619, -0.042803, 0.282745, 0.153629, 0.939992, 0.588169, 0.189058, 0.461301, -0.667901, -0.314791}
  #declare cdata = array[16];
  #debug "\n\nData\n"
  #for(i,0,dimension_size(data,1)-1)
    #declare cdata[i] = R2C(data[i]);
    #debug concat(Cstr(cdata[i]), "\n")
  #end  

  #debug "\n\nFFT\n"  
  FFT(cdata)
  CdebugArr(cdata)
 
  // power
  #debug "\nPower\n"
  #for(i,0,dimension_size(cdata,1)-1)
    #debug concat(str(cdata[i].x * cdata[i].x + cdata[i].y * cdata[i].y,0,-1), "\n")  
  #end
  
  // phase
  #debug "\nPhase\n"
  #for(i,0,dimension_size(cdata,1)-1)
    #debug concat(str(atan2(cdata[i].x, cdata[i].y).x,0,-1), "\n")  
  #end

  #debug "\nIFFT\n"  
  IFFT(cdata)
  CdebugArr(cdata)
  #debug "\n"

  // set up array of arrays, 2D data
  #declare arr2d = array[4]{
    array[8]{1, 2, 3, 4, 5, 6, 7, 8},
    array[8]{1, 2, 3, 4, 5, 6, 7, 8},
    array[8]{1, 2, 3, 4, 5, 6, 7, 8},
    array[8]{1, 2, 3, 4, 5, 6, 7, 8}
  };
  // turn real into complex
  #declare data = array[4];
  #for(row, 0, dimension_size(arr2d,1)-1)
    #local data[row] = array[dimension_size(arr2d[row],1)];
    #for(col, 0, dimension_size(arr2d[row],1)-1)    
      #local data[row][col] = R2C(arr2d[row][col]);
    #end
  #end  
  #debug "\n2D FFT Data\n"
  Cdebug2Darr(data)
  
  #debug "\n2D FFT\n"  
  FFT2D(data)
  Cdebug2Darr(data)
  
  #debug "\n2D IFFT\n"  
  IFFT2D(data)
  Cdebug2Darr(data)
  #debug "\n"
#end

#version FFT_Inc_Temp;
#end
